#include "coordinate_system.h"
#include <QPainterPath>        
#define steps 50

int binomialCoefficient(int n, int k) {
    int res = 1;
    if (k > n - k)
        k = n - k;
    for (int i = 0; i < k; ++i) {
        res *= (n - i);
        res /= (i + 1);
    }
    return res;
}
#define bernsteinPolynomial(n,i,t)(binomialCoefficient(n, i) * pow(t, i) * pow(1 - t, n - i))

void CoordinateSystem::finalize() {
  layout->addWidget(addb,fields.size()+1,0);
  layout->addWidget(clearb,fields.size()+1,1);
  layout->addWidget(button,fields.size()+2,0,1,2);
  container->setLayout(layout);
  iIndex->setRange(0, fields.size() - 1);
}
void CoordinateSystem::updateResultText() {
  resultLabel->clear();
   QString resultText;
   #define n fields.size() - 1
   for (double t = tStart->value(); t <= tEnd->value(); t += tabIndex->value()) {
       double value = bernsteinPolynomial(n, iIndex->value(), t);
       resultText += QString("t = %1, B_%2^%3(t) = %4\n").arg(t).arg(n).arg(iIndex->value()).arg(value);
   }
   resultLabel->setText(resultText);
}
void CoordinateSystem::updateStart() {
    if (tStart->value() > tEnd->value())
        tEnd->setValue(tStart->value());
    updateResultText();
}
void CoordinateSystem::updateEnd() {
    if (tEnd->value() < tStart->value())
        tStart->setValue(tEnd->value());
    updateResultText();
}
CoordinateSystem::CoordinateSystem(QWidget *parent) : QWidget(parent) {
    layout = new QGridLayout(this);
    adddialog = new QPushButton();
    text = new QLabel("Відображати координати кожні                    точок", this);
    text2 = new QLabel("Червоні точки - опорні, білі - керуючі. Після відмальовки крива лишається", this);
    text2->setAlignment(Qt::AlignRight);
        displayus = new QSpinBox(this);
        displayus->setRange(2,step);
        displayus->setValue(step/2);
        connect(displayus, QOverload<int>::of(&QSpinBox::valueChanged), this, &CoordinateSystem::up);
    QLabel* x = new QLabel("X", this);
    QLabel* y = new QLabel("Y", this);
    layout->addWidget(x, 0, 0);
    layout->addWidget(y, 0, 1);
    for (char i = 0; i < 3; i++)add();
    button = new QPushButton("Draw", this);    
    clearb = new QPushButton("Clear", this);
    addb = new QPushButton("Add", this);
    connect(button, &QPushButton::clicked, this, &CoordinateSystem::onButtonClick);
    connect(addb, &QPushButton::clicked, this, &CoordinateSystem::addandfix);
    connect(clearb, &QPushButton::clicked, this, &CoordinateSystem::clear);
    clearb->setFixedWidth(60);
    addb->setFixedWidth(60);
    container = new QWidget(this);
    QLabel* tLabel = new QLabel("Діапазон t", this);
    tStart = new QDoubleSpinBox(this);
    tStart->setRange(0, 1);
    tStart->setSingleStep(0.01);
    tEnd = new QDoubleSpinBox(this);
    tEnd->setRange(0, 1);
    tEnd->setSingleStep(0.01);
    tStart->setValue(0);
    tEnd->setValue(100);
    QLabel* iLabel = new QLabel("Індекс i", this);
    iIndex = new QSpinBox(this);    
    QLabel* tabLabel = new QLabel("Крок", this);
    tabIndex = new QDoubleSpinBox(this);
    tabIndex->setRange(0.1, 1);
    tabIndex->setSingleStep(0.05);
    tabIndex->setValue(0.2);
    iIndex->setValue(0);
    layout->addWidget(tLabel, 0, 2);
    layout->addWidget(tStart, 0, 3);
    layout->addWidget(tEnd, 0, 4);
    layout->addWidget(iLabel, 1, 2);
    layout->addWidget(iIndex, 1, 3);
    layout->addWidget(tabLabel, 2, 2);
    layout->addWidget(tabIndex, 2, 3);
    finalize();
    resultLabel = new QLabel(this);
        resultLabel->setAlignment(Qt::AlignTop | Qt::AlignLeft);
        resultLabel->setWordWrap(true);  // Якщо виведення буде довгим    
        resultLabel->setGeometry(138,114,762,336);
        connect(tabIndex, QOverload<double>::of(&QDoubleSpinBox::valueChanged), this, &CoordinateSystem::updateResultText);
        connect(iIndex, QOverload<int>::of(&QSpinBox::valueChanged), this, &CoordinateSystem::updateResultText);
        connect(tStart, QOverload<double>::of(&QDoubleSpinBox::valueChanged), this, &CoordinateSystem::updateStart);
        connect(tEnd, QOverload<double>::of(&QDoubleSpinBox::valueChanged), this, &CoordinateSystem::updateEnd);
        updateResultText();
}
void CoordinateSystem::add() {
    fields.emplace_back();
    for (char j = 0; j < 2; j++) {
        fields.back()[j] = new QSpinBox(this);
        fields.back()[j]->setFixedWidth(60);
        fields.back()[j]->setFixedHeight(30);
        fields.back()[j]->setSingleStep(10);
        fields.back()[j]->setRange(-centerX, centerY);
        fields.back()[j]->setValue(10 * (rand() % 40 - 20));
        layout->addWidget(fields.back()[j], fields.size(), j);
        connect(fields.back()[j], QOverload<int>::of(&QSpinBox::valueChanged), this, &CoordinateSystem::up);
    }
}

void CoordinateSystem::addandfix() {
    add();
    finalize();
    update();
}

void CoordinateSystem::up() {update();}
void CoordinateSystem::clear() {dots.clear();update();}
void CoordinateSystem::paintEvent(QPaintEvent*) {
  displayus->setGeometry(212,height()-30,60,30);
  text->setGeometry(10,0,height(),height()*2-30);
  text2->setGeometry(width()-510,0,height(),height()*2-30);
    setMinimumHeight(fields.size() * 36 + 150);
    container->adjustSize();
    bool drawn = false;
    painter = new QPainter(this);
    painter->setRenderHint(QPainter::Antialiasing);
    painter->drawLine(0, centerY, width(), centerY);
    painter->drawLine(centerX, 0, centerX, height());
    painter->drawLine(centerX - 5, 11, centerX, 0);
    painter->drawLine(centerX + 5, 11, centerX, 0);
    painter->drawLine(width()-11, centerY - 5, width(), centerY);
    painter->drawLine(width()-11, centerY + 5, width(), centerY);
    for (short i = centerX; i < width(); i += step)  
        painter->drawLine(i, centerY - 5, i, centerY + 5);
    for (short i = centerX; i > 0; i -= step)        
        painter->drawLine(i, centerY - 5, i, centerY + 5);
    
    for (short i = centerY; i < height(); i += step) {
        painter->drawLine(centerX - 5, i, centerX + 5, i);
            painter->drawText(centerX + 5, centerY - 5 + step, QString::number(-step));
        drawn = true;
    }

    for (short i = centerY; i > 0; i -= step) 
        painter->drawLine(centerX - 5, i, centerX + 5, i);
    short newMin1 = -centerX;
    for (char j = 0; j < fields.size(); j++) {
                fields[j][0]->blockSignals(true);
                fields[j][0]->setRange(newMin1, -newMin1);
                fields[j][0]->blockSignals(false);
    }
    short newMin = centerY;
    if (fields[0][1]->maximum() != newMin)
    for (char j = 0; j < fields.size(); j++) {
                fields[j][1]->blockSignals(true);
                fields[j][1]->setRange(-newMin,newMin);
                fields[j][1]->blockSignals(false);
    }
    painter->setPen(Qt::green);
        for (std::vector<short>&paste : dots)
        for (char i=2;i<paste.size();i+=2)
      painter->drawLine(centerX+paste[i-2], centerY-paste[i-1], centerX+paste[i], centerY-paste[i+1]);
        painter->setPen(Qt::red);
        for (std::vector<short>&paste : dots) {
            QPainterPath path;
            path.moveTo(centerX+paste[0], centerY-paste[1]);
            for (unsigned char step = 1; step <= steps; step++) {
                float t = (float)step/steps;
                std::vector<QPointF> temp;
                
                for (size_t i = 0; i < paste.size(); i += 2)
                    temp.emplace_back(centerX + paste[i], centerY - paste[i + 1]);
        
                while (temp.size()) {
                    for (size_t i = 0; i < temp.size() - 1; i++) {
                        temp[i].setX((1 - t) * temp[i].x() + t * temp[i + 1].x());
                        temp[i].setY((1 - t) * temp[i].y() + t * temp[i + 1].y());
                    }
                    temp.pop_back();
                }
        
                path.lineTo(temp[0]);
                if(!((step-(displayus->value()/2))%displayus->value())){painter->setPen(Qt::white);
                painter->drawText(temp[0].x()+5,temp[0].y()-4, QString::number(temp[0].x()-centerX)+", "+QString::number(centerY-temp[0].y()));
                painter->setPen(Qt::blue);
                painter->drawEllipse(temp[0], 4,4);
                painter->setPen(Qt::red);}
            }
        
            painter->drawPath(path);
        }
        
    painter->setPen(Qt::white);
    painter->drawText(centerX - step + 5, centerY - 5, QString::number(-step));
    painter->drawText(centerX + 5, centerY - 5 - step, ' ' + QString::number(step));
    painter->drawText(centerX + 5, 15, "Y");
    painter->drawText(width() - 20, centerY - 7, "X");
    painter->drawText(centerX + 5, centerY - 5, "0");

    painter->drawText(centerX + step + 5, centerY - 5, QString::number(step));
    for (char i=0;i<fields.size();i++){
        if(!i||i==fields.size()-1)painter->setPen(Qt::red);
    painter->drawEllipse(centerX + fields[i][0]->value()-5, centerY - fields[i][1]->value()-5, 11, 11);
    painter->setPen(Qt::white);
    painter->drawText(centerX + fields[i][0]->value()+2, centerY - fields[i][1]->value()-10, QString::number(i));
    }
    painter->end();
}

void CoordinateSystem::onButtonClick() {
  std::vector<short> paste;
  for (unsigned char i=0;i<fields.size()*2;i++)
  paste.push_back(fields[i/2][i%2]->value());
  dots.push_back(paste);
  update();
}
